/*********************************************
 * Author: Guido Socher
 * Copyright: GPL V2
 * http://www.gnu.org/licenses/gpl.html
 *
 * Based on the enc28j60.c file from the AVRlib library by Pascal Stang.
 * For AVRlib See http://www.procyonengineering.com/
 * Used with explicit permission of Pascal Stang.
 *
 * Title: Microchip ENC28J60 Ethernet Interface Driver
 * Chip type           : ATMEGA88 with ENC28J60
 *********************************************/
#include <stdlib.h>
#include <string.h>
#include <avr/io.h>
#include "avr_compat.h"
#include "enc28j60.h"
#include "timeout.h"

#define F_CPU 16000000UL

#ifndef ALIBC_OLD
#include <util/delay.h>
#else
#include <avr/delay.h>
#endif

static uint8_t Enc28j60Bank;

static uint16_t NextPacketPtr;
#define ENC28J60_CONTROL_PORT   PORTC
#define ENC28J60_CONTROL_DDR    DDRC
#define ENC28J60_CONTROL_CS     1

// Set CS to 0 = active.
#define CSACTIVE ENC28J60_CONTROL_PORT&=~(1<<ENC28J60_CONTROL_CS)
// Set CS to 1 = passive.
#define CSPASSIVE ENC28J60_CONTROL_PORT|=(1<<ENC28J60_CONTROL_CS)

#define waitspi() while(!(SPSR&(1<<SPIF)))

uint16_t GetNextPointer(uint16_t P)
{
	if (P == RXSTOP_INIT)
	{
		return 0;
	}
	else
	{
		return P + 1;
	}
}


uint16_t GetPrevPointer(uint16_t P)
{
	if (P == RXSTART_INIT)
	{
		return RXSTOP_INIT;
	}
	else
	{
		return P - 1;
	}
}

// Issues a read command with opcode 'op' and 'address' to the ENC28J60 chip.
//
// 'op' is shifted so that the opcode is on the 3 most significant bits.
// 'address' contains the address of the control register on the 5 least
// significant bits and the 7th bit indicates wether a dummy byte should
// be sent or not.
uint8_t enc28j60ReadOp(uint8_t op, uint8_t address)
{
	CSACTIVE;

    // issue read command
    SPDR = op | (address & ADDR_MASK);
    waitspi();

    // read data
    SPDR = 0x00;
    waitspi();
    // do dummy read if needed (for mac and mii, see datasheet page 29)

    if(address & 0x80)
    {
        SPDR = 0x00;
        waitspi();
    }
	
	uint8_t rval = SPDR;

    // release CS
    CSPASSIVE;
    return rval;

}

// Issues a write command with opcode 'op' and 'address' wich will write
// byte 'data' to the ENC28J60 chip.
//
// 'op' is shifted so that the opcode is on the 3 most significant bits.
// 'address' contains the address of the control register on the 5 least
// significant bits.
void enc28j60WriteOp(uint8_t op, uint8_t address, uint8_t data)
{
	CSACTIVE;
		
	SPDR = (op & 0xe0) | (address & 0x1f);
	waitspi();

	SPDR = data;
	waitspi();

	CSPASSIVE;
}

void enc28j60ReadBuffer(uint16_t len, uint8_t* data)
{
    CSACTIVE;

    // Issue read command.
    SPDR = ENC28J60_READ_BUF_MEM;
    waitspi();
    while(len)
    {
        len--;
        // read data
        SPDR = 0x00;
        waitspi();
        *data = SPDR;
        data++;
    }
    *data='\0';
    CSPASSIVE;
}

void enc28j60WriteBuffer(uint16_t len, uint8_t* data)
{
    CSACTIVE;
    // issue write command
    SPDR = ENC28J60_WRITE_BUF_MEM;
    waitspi();
    while(len)
    {
        len--;
        // write data
        SPDR = *data;
        data++;
        waitspi();
    }
    CSPASSIVE;
}

void enc28j60SetBank(uint8_t address)
{
    // set the bank (if needed)
    if((address & BANK_MASK) != Enc28j60Bank)
    {
        // set the bank
        enc28j60WriteOp(ENC28J60_BIT_FIELD_CLR, ECON1, (ECON1_BSEL1|ECON1_BSEL0));
        enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, ECON1, (address & BANK_MASK)>>5);
        Enc28j60Bank = (address & BANK_MASK);
    }
}

uint8_t enc28j60Read(uint8_t address)
{
    // set the bank
    enc28j60SetBank(address);
    // do the read
    return enc28j60ReadOp(ENC28J60_READ_CTRL_REG, address);
}

void enc28j60Write(uint8_t address, uint8_t data)
{
    // set the bank
    enc28j60SetBank(address);
    // do the write
    enc28j60WriteOp(ENC28J60_WRITE_CTRL_REG, address, data);
}

void enc28j60PhyWrite(uint8_t address, uint16_t data)
{
    // set the PHY register address
    enc28j60Write(MIREGADR, address);
    // write the PHY data
    enc28j60Write(MIWRL, data);
    enc28j60Write(MIWRH, data>>8);
    // wait until the PHY write completes
    while(enc28j60Read(MISTAT) & MISTAT_BUSY){
        _delay_us(15);
    }
}


void enc28j60Init(uint8_t* macaddr)
{
    // initialize I/O
    ENC28J60_CONTROL_DDR |= 1<<ENC28J60_CONTROL_CS;
    CSPASSIVE;
    //
    DDRB  |= 1<<PB7 |1<<PB4 | 1<<PB5; // mosi, sck, ss output
    cbi(DDRB,PINB6); // MISO is input
    //
    CSPASSIVE;
    cbi(PORTB,PB5); // MOSI low
    cbi(PORTB,PB7); // SCK low
    //
    // initialize SPI interface
    // master mode and Fosc/2 clock:
    SPCR = (1<<SPE)|(1<<MSTR) ;//|(1 << SPR1) |(1 << SPR0);
    SPSR |= (1<<SPI2X);
    //SPSR &= (~(1<<SPI2X));

    // perform system reset
    enc28j60WriteOp(ENC28J60_SOFT_RESET, 0, ENC28J60_SOFT_RESET);
    delay_ms(200);

    // check CLKRDY bit to see if reset is complete
    // The CLKRDY does not work. See Rev. B4 Silicon Errata point. Just wait.
    //while(!(enc28j60Read(ESTAT) & ESTAT_CLKRDY));
    // do bank 0 stuff
    // initialize receive buffer
    // 16-bit transfers, must write low byte first
    // set receive buffer start address
    NextPacketPtr = RXSTART_INIT;
	NextPacketPtr = GetPrevPointer(NextPacketPtr);
    // Rx start
    enc28j60Write(ERXSTL, RXSTART_INIT&0xFF);
    enc28j60Write(ERXSTH, RXSTART_INIT>>8);
    // set receive pointer address
    enc28j60Write(ERXRDPTL, NextPacketPtr&0xFF);
    enc28j60Write(ERXRDPTH, NextPacketPtr>>8);
    // RX end
    enc28j60Write(ERXNDL, RXSTOP_INIT&0xFF);
    enc28j60Write(ERXNDH, RXSTOP_INIT>>8);
    // TX start
    enc28j60Write(ETXSTL, TXSTART_INIT&0xFF);
    enc28j60Write(ETXSTH, TXSTART_INIT>>8);
    // TX end
    enc28j60Write(ETXNDL, TXSTOP_INIT&0xFF);
    enc28j60Write(ETXNDH, TXSTOP_INIT>>8);

    // do bank 1 stuff, packet filter:

    enc28j60Write(ERXFCON, ERXFCON_UCEN|ERXFCON_CRCEN|ERXFCON_BCEN);

    // do bank 2 stuff
    // enable MAC receive
    enc28j60Write(MACON1, MACON1_MARXEN|MACON1_TXPAUS|MACON1_RXPAUS);
    // bring MAC out of reset
    enc28j60Write(MACON2, 0x00);
    // enable automatic padding to 60bytes and CRC operations
    enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, MACON3, MACON3_PADCFG0|MACON3_TXCRCEN|MACON3_FRMLNEN|MACON3_FULDPX|MACON3_HFRMLEN);
    // set inter-frame gap (non-back-to-back)
    enc28j60Write(MAIPGL, 0x12);
    enc28j60Write(MAIPGH, 0x0C);
    // set inter-frame gap (back-to-back)
    enc28j60Write(MABBIPG, 0x12);
    // Set the maximum packet size which the controller will accept
    // Do not send packets longer than MAX_FRAMELEN:
    enc28j60Write(MAMXFLL, MAX_FRAMELEN&0xFF);
    enc28j60Write(MAMXFLH, MAX_FRAMELEN>>8);
    // do bank 3 stuff
    // write MAC address
    // NOTE: MAC address in ENC28J60 is byte-backward

    enc28j60Write(MAADR5, macaddr[0]);
    enc28j60Write(MAADR4, macaddr[1]);
    enc28j60Write(MAADR3, macaddr[2]);
    enc28j60Write(MAADR2, macaddr[3]);
    enc28j60Write(MAADR1, macaddr[4]);
    enc28j60Write(MAADR0, macaddr[5]);

    // no loopback of transmitted frames
    enc28j60PhyWrite(PHCON2, PHCON2_HDLDIS);
    // switch to bank 0
    enc28j60SetBank(ECON1);
    // enable interrutps
//        enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, EIE, EIE_INTIE|EIE_PKTIE);
    // enable packet reception
    enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, ECON1, ECON1_RXEN);

}

// read the revision of the chip:
uint8_t enc28j60getrev(void)
{
    return(enc28j60Read(EREVID));
}

void enc28j60PacketSend(uint16_t len, uint8_t* packet)
{
    // Set the write pointer to start of transmit buffer area
    enc28j60Write(EWRPTL, (TXSTART_INIT + 1)&0xFF);
    enc28j60Write(EWRPTH, (TXSTART_INIT + 1)>>8);
    // Set the TXND pointer to correspond to the packet size given
    enc28j60Write(ETXNDL, (TXSTART_INIT+len)&0xFF);
    enc28j60Write(ETXNDH, (TXSTART_INIT+len)>>8);
    // write per-packet control byte (0x00 means use macon3 settings)
    enc28j60WriteOp(ENC28J60_WRITE_BUF_MEM, 0, 0x00);
    // copy the packet into the transmit buffer
    enc28j60WriteBuffer(len, packet);
    // send the contents of the transmit buffer onto the network
    enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, ECON1, ECON1_TXRTS);
    // Reset the transmit logic problem. See Rev. B4 Silicon Errata point 12.
    if( (enc28j60Read(EIR) & EIR_TXERIF) ){
            enc28j60WriteOp(ENC28J60_BIT_FIELD_CLR, ECON1, ECON1_TXRTS);
    }
}

// Gets a packet from the network receive buffer, if one is available.
// The packet will by headed by an ethernet header.
//      maxlen  The maximum acceptable length of a retrieved packet.
//      packet  Pointer where packet data should be stored.
// Returns: Packet length in bytes if a packet was retrieved, zero otherwise.
uint16_t enc28j60PacketReceive(uint16_t maxlen, uint8_t* packet)
{
    uint16_t rxstat;
    uint16_t len;
    uint16_t tnp;
    uint16_t pktcount;
	int i;

    // check if a packet has been received and buffered
    //if( !(enc28j60Read(EIR) & EIR_PKTIF) ){
    // The above does not work. See Rev. B4 Silicon Errata point 6.
	pktcount = enc28j60Read(EPKTCNT);
    if( pktcount ==0 ){
        return(0);
    }

    // Set the read pointer to the start of the received packet
	
	uint8_t tempbuff[7];
	
	for (i = 0;i < 7;i += 1)
	{
		enc28j60Write(ERDPTL, (NextPacketPtr));
		enc28j60Write(ERDPTH, (NextPacketPtr)>>8);
		
		NextPacketPtr = GetNextPointer(NextPacketPtr);
		
		enc28j60ReadBuffer(1,tempbuff + i);
	}
	
	tnp = *((uint16_t *)(tempbuff + 1));
	len = *((uint16_t *)(tempbuff + 3));
	rxstat = *((uint16_t *)(tempbuff + 5));

    // limit retrieve length
    if (len>maxlen-1){
        len=maxlen-1;
    }

	for (i = 0;i < len;i += 1)
	{
		enc28j60Write(ERDPTL, (NextPacketPtr));
        enc28j60Write(ERDPTH, (NextPacketPtr)>>8);

		NextPacketPtr = GetNextPointer(NextPacketPtr);

		enc28j60ReadBuffer(1,packet + i);
	}

    // Move the RX read pointer to the start of the next received packet
    // This frees the memory we just read out
	NextPacketPtr = GetPrevPointer(NextPacketPtr);
    enc28j60Write(ERXRDPTL, (NextPacketPtr));
    enc28j60Write(ERXRDPTH, (NextPacketPtr)>>8);
    // decrement the packet counter indicate we are done with this packet
    enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, ECON2, ECON2_PKTDEC);

    return len - 4;
}
